function dataFVS = FVSmod(kernfunc, X, nbFV, maxFitness)
%--------------------------------------------------------------------------------------------
% Feature vector selection using the a priori scheme to select the feature vectors (FV).
% The FVs are choosen to minimize in F the reconstruction error of all the input data.
% This selection procedure ends as soon as we reach an numerical basis in F, hit 
% the maximum fitness value maxFitness (0 up to 1),or exhausted the amount of FV allowed
% by nbFV (1 up to the size of the input database (M)).
%
% INPUT
%   X           : is the data where each row is a given sample.
%   kernfunc    : is handle to the kernel function
%                 (see gausskernel.m and polykernel.m)
%   nbFV        : if omitted the size of the data X is the default value.
%   maxFitness  : if omitted the default value will be 1-eps.
%
% Return: 
%   dataFVS     : structure carring the selected FV, their numbers and the 
%                 projected input data into the FV subspace
%--------------------------------------------------------------------------------------------
% Copyright  Gaston Baudat & Fatiha Anouar September 2001. Exton PA 19341 USA
% gaston.baudat@eu.effem.com, fatiha.anouar@effem.com
%--------------------------------------------------------------------------------------------
% Modifications:
% David Brown & Gaston Baudat 10/14/2001
%
% David Brown has provided us improved Matlab code to speed up the procedure (9/2001)
% His initial contribution was optimized for Gauss's kernel.
% On the other end the current version can handle other kernels.
% To know more about his version you can contact him at: BRWDAV010@mail.uct.ac.za
%
% This version evaluate the whole kernel matrix K before starting, this is faster but you must be aware that
% you need to support enough memory to store it.
% Some "for" loops have been replaced by a more efficient vectorial approach (D. Brown)
% The inversion of the Kss matrix is done now recursively (assuming symetric matrix), this allows to check its
% numerical condition (rank estimation), while the inversion complexity drops from M^3 to M^2 (G. Baudat & F. Anouar)
% See invrec.m file for more information.
%
% Modifications:
% Ante Jukic 10/02/2013
%
% Added input kernfunc.
%

% variables, basic initialisations & test for validity of arguments.


[M,N]=size(X);  % M number of sample, N input space size.

switch nargin 
    case 2
          nbFV = M;
          maxFitness = 1-eps;  
    case 3
          maxFitness=1-eps;
    case 4
    otherwise
          fprintf('Error: function called with a wrong number of parameter. Allowed 2, 3, or 4');
end

if nbFV>M
    nbFV=M;
end

if nbFV<=0
    nbFV=1;
end

if (maxFitness<0)|(maxFitness>1)
    maxFitness=1-eps;
end

L = 0;            % current size of the FV set
S = [];           % current FV set (list of the sample index in X)
R = 1:M;          % current remaining vector (not selected in S)
baseFound = 0;    % flag set to one if a numerical base was found (the S set is completed)
fitness = 0;      % current fitness value


% initialize kernel
% p1 = 2;
% K = polykernel(X',X',p1);
K = kernfunc(X',X');

% compute the module of each sample into F

modSample=zeros(1,M);

for I=1:M
    modSample(I)=1/sqrt(K(I,I));
end

% search for the first feature vector using the gobal fitness function
% each input vector is taken, one after an other, as a candidate.
% The best candidate is the first feature vector of the set S
% Ksx is a 1xM matrix where each column is a Ksi vector for the ith input vector

L=1;
for I=1:M                                   % look for a new candidate
    Kss=K(I,I);							    % update the Kss value (matrix 1x1)
    if abs(Kss)>eps                         % Kss is numericaly invertible
        Ksx(L,:)=K(I,:).*modSample;  		% update Ksx matrix, each column is a Ksi vector
        Js=sum(Ksx.*Ksx/Kss)/M; 			% compute the global fitness
        if Js>fitness                       % check if the candidate is the best so far
            fitness=Js;
            newFV=I;                        
        end
    end
end
 
% main loop to selected the other FV, stop when S is a numerical base, maxFitness is reached or nbFV is reached.

S(1)=newFV;                 % initialize the set S with the first selected FV (see above) 
R(newFV)=0;
invKss=[];
Ksx=[];
Js=0;

% Basic algorithm description
% ...........................
% for a given set S we compute the global fitness
% this is done input vector after input vector in order to find the vector having the lowest
% (local) fitness in same time. Then this one will be the next candiate as FV.
% That approach speeds up considerably the procedure while selectioning, at each iteration, as
% a new FV the "most orthogonal" at the previous set S.
% Some improvements of the Matlab code have been made by David Brown (9/2001) 

while (L<nbFV)&(Js<maxFitness)&(~baseFound)
    newFV=0;
    [invKss,r]=invrec(invKss,K(S(L),S));   % recursive inversion of the Kss (adding a new line/column, must be symetric)
	if L==r             				   % test if the current Kss matrix is well numerically conditioned
        Ksx=[Ksx, K(:,S(L)).*modSample'];  % update the Ksx matrix with a new FV, each column is a Ksi vector
        J=sum(Ksx.*(Ksx*invKss),2);		   % compute all the local fitnesses
        Js=(sum(J(R~=0))+L)/M;             % compute the global fitness with the none selected vectors
        [fitnessLocal,newFV]=min(J);       % find new FV 
        if fitnessLocal>=(1-eps)           % just in case we are adding an unnecessary FV
            newFV=0;
        end
	else
		baseFound=1;
    end
    if newFV==0                     % base found, keep the previous FV set S and Ksi matrix
		baseFound=1;
        fprintf('base found %d\n',L);
    else
        if Js<maxFitness            % not yet reached the maximum of the global fitness argument
            fprintf('Iteration: %d current global fitness = %f\n',L,Js);
            R(newFV)=0;             % remove the new FV from the remaining sample list    
            L=L+1;
            S(L)=newFV;             % add the new FV to the S set
        end
    end
end
fprintf('Last iteration: %d current global fitness = %f\n',L,Js);
fprintf('list of FV number: ');    % print the FV set
fprintf('%d ',S);
fprintf('\n');

% Projection all the input data to the subspace of the FV into F (explicit projection using the kernel)
P=K(:,S);%fvskern(ker,X(S,:),X);

% returned structure
dataFVS=struct('FV',X(S,:),'S',S, 'P', P);
